from __future__ import annotations

import functools
import itertools
import random
from collections.abc import Iterator

from musiclib import config
from musiclib.card import Card
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet
from musiclib.util.cache import Cached


class Chord(NoteSet):
    intervals_to_name = {
        # triads
        frozenset({0, 4, 7}): 'major',
        frozenset({0, 3, 7}): 'minor',
        frozenset({0, 3, 6}): 'diminished',
        # 7th
        frozenset({0, 4, 7, 11}): 'maj7',
        frozenset({0, 4, 7, 10}): '7',
        frozenset({0, 3, 7, 10}): 'min7',
        frozenset({0, 3, 6, 10}): 'half-dim7',
        frozenset({0, 3, 6, 9}): 'dim7',
        # 6th
        frozenset({0, 4, 7, 9}): '6',
        frozenset({0, 3, 7, 9}): 'm6',
        # etc
        frozenset({0, 4, 8}): 'aug',
        frozenset({0, 2, 7}): 'sus2',
        frozenset({0, 5, 7}): 'sus4',
    }
    name_to_intervals = {v: k for k, v in intervals_to_name.items()}
    root: Note
    name: str

    def __init__(self, notes: frozenset[Note], *, root: Note):
        if root is None:
            raise TypeError('Chord requires root note. Use NoteSet if there is no root')
        super().__init__(notes, root=root)

    def _repr_html_(
        self,
        html_classes: tuple[str, ...] = ('card',),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
    ) -> str:
        return self.repr_card(
            html_classes=html_classes,
            title=title or f'{self.root.name} {self.name}',
            subtitle=subtitle,
            header_href=header_href,
            background_color=background_color,
            piano_html=self.to_piano_image(),
        )


class SpecificChord(Cached, Card):
    def __init__(
        self,
        notes: frozenset[SpecificNote],
        *,
        root: str | Note | None = None,
    ):
        if not isinstance(notes, frozenset):
            raise TypeError(f'expected frozenset, got {type(notes)}')

        if isinstance(root, str):
            root = Note(root)

        notes_abstract = SpecificNote.to_abstract(notes)
        if root is not None and root not in notes_abstract:
            raise KeyError('root should be one of notes')

        self.notes = notes
        self.root = root
        self.abstract = Chord(notes_abstract, root=root) if root is not None else NoteSet(notes_abstract)
        self.root_specific = frozenset(note for note in notes if note.abstract == root) if root is not None else frozenset()

        self.notes_ascending = tuple(sorted(notes))
        self.intervals = tuple(note - self.notes_ascending[0] for note in self.notes_ascending)  # from lowest note
        self.key = self.notes, self.root

    @classmethod
    def random(cls, n_notes: int | None = None, octaves: tuple[int, ...] = (3, 4, 5)) -> SpecificChord:
        if n_notes is None:
            n_notes = random.randint(2, 5)
        notes_space = tuple(
            SpecificNote(note, octave)
            for note, octave in itertools.product(config.chromatic_notes, octaves)
        )
        notes = frozenset(random.sample(notes_space, n_notes))
        return cls(notes)

    @classmethod
    def from_str(cls, string: str) -> SpecificChord:
        notes_, _, root_ = string.partition('/')
        root = Note(root_) if root_ else None
        notes_2 = notes_.split('_')
        if len(notes_2) != len(set(notes_2)):
            raise NotImplementedError('SpecificChord_s with non unique notes are not supported')
        notes = frozenset(SpecificNote.from_str(note) for note in notes_2)
        return cls(notes, root=root)

    def notes_combinations(self) -> Iterator[tuple[SpecificNote, SpecificNote]]:
        yield from itertools.combinations(self.notes_ascending, 2)

    def find_intervals(self, interval: int) -> tuple[tuple[SpecificNote, SpecificNote], ...]:
        return tuple((n, m) for n, m in self.notes_combinations() if abs(m - n) == interval)

    def __len__(self) -> int:
        return len(self.notes)

    def __getitem__(self, item: int) -> SpecificNote:
        return self.notes_ascending[item]

    def __iter__(self) -> Iterator[SpecificNote]:
        return iter(self.notes_ascending)

    def __repr__(self) -> str:
        x = '_'.join(repr(note) for note in self.notes_ascending)
        if self.root is not None:
            return f'{x}/{self.root.name}'
        return x

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SpecificChord):
            return NotImplemented
        return self.key == other.key

    def __hash__(self) -> int:
        return hash(self.key)

    def __sub__(self, other: SpecificChord) -> int:
        return sum(abs(a - b) for a, b in zip(self, other, strict=True))

    def __add__(self, other: int) -> SpecificChord:
        """transpose"""
        if not isinstance(other, int):
            raise TypeError('only adding integers is allowed (transposition)')
        root = self.root + other if self.root is not None else None
        return SpecificChord(frozenset(note + other for note in self), root=root)

    @functools.cached_property
    def transposed_to_C0(self) -> SpecificChord:
        return self + (SpecificNote('C', 0) - self[0])

    def to_piano_image(self) -> str:
        from musiclib.noterange import NoteRange
        from musiclib.piano import Piano
        noterange = NoteRange(self[0], self[-1]) if self.notes else None
        return Piano(
            note_colors=dict.fromkeys(self.notes, config.RED),
            squares={note: {'text': str(note), 'text_size': '8'} for note in self},
            noterange=noterange,
        )._repr_svg_()

    def _repr_html_(
        self,
        html_classes: tuple[str, ...] = ('card',),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
    ) -> str:
        return self.repr_card(
            html_classes=html_classes,
            title=title or repr(self),
            subtitle=subtitle,
            header_href=header_href,
            background_color=background_color,
            piano_html=self.to_piano_image(),
        )

    def __getnewargs_ex__(self) -> tuple[tuple[frozenset[SpecificNote]], dict[str, Note | None]]:
        return (self.notes,), {'root': self.root}



from collections import defaultdict

from musiclib.chord import SpecificChord
from musiclib.note import SpecificNote
from musiclib.noterange import NoteRange
from musiclib.noteset import NoteSet

SpecificChordGraph = dict[SpecificChord, frozenset[SpecificChord]]
AbstractChordGraph = dict[NoteSet, frozenset[NoteSet]]


class Transition:
    def __init__(self, a: SpecificChord, b: SpecificChord):
        self.a = a
        self.b = b

    @staticmethod
    def arrow(a: SpecificNote, b: SpecificNote) -> str:
        if a < b:
            return '︎\\'
        elif a == b:
            return '|'
        else:
            return '/'

    def __repr__(self) -> str:
        return '\n'.join((
            str(self.a),
            '  '.join(self.arrow(na, nb) for na, nb in zip(self.a, self.b, strict=True)),
            str(self.b),
        ))


def chord_transitions(
    chord: SpecificChord,
    noterange: NoteRange,
    unique_abstract: bool = False,
    same_length: bool = True,
) -> frozenset[SpecificChord]:
    out = set()
    for note in chord:
        for add in (-1, 1):
            if (new_note := noterange.noteset.add_note(note, add)) not in noterange:
                continue
            notes = chord.notes - {note} | {new_note}
            if same_length and len(notes) != len(chord.notes):
                continue
            if unique_abstract and len(notes) > len({n.abstract for n in notes}):
                continue
            out.add(SpecificChord(notes))
    return frozenset(out)


def transition_graph(
    start_chord: SpecificChord,
    noterange: NoteRange,
    unique_abstract: bool = False,
    same_length: bool = True,
) -> dict[SpecificChord, frozenset[SpecificChord]]:
    graph: defaultdict[SpecificChord, set[SpecificChord]] = defaultdict(set)

    def _graph(chord: SpecificChord) -> None:
        if chord in graph:
            return
        childs = chord_transitions(chord, noterange, unique_abstract, same_length)
        graph[chord] |= childs
        for child in childs:
            _graph(child)

    _graph(start_chord)
    return {k: frozenset(v) for k, v in graph.items()}


def abstract_graph(g: SpecificChordGraph) -> AbstractChordGraph:
    graph: defaultdict[NoteSet, set[NoteSet]] = defaultdict(set)

    for k, v in g.items():
        graph[k.abstract] |= {c.abstract for c in v}

    return {k: frozenset(v) for k, v in graph.items()}



from __future__ import annotations

import itertools
import random
from collections.abc import Iterator
from typing import TypeVar
from typing import no_type_check
from typing import overload

import pipe21 as P

from musiclib import config
from musiclib.card import Card
from musiclib.config import RED
from musiclib.note import AnyNote
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.note import str_to_note
from musiclib.util import typeguards
from musiclib.util.cache import Cached


@no_type_check
def bits_to_intervals(bits: str) -> frozenset[int]:
    return (
        bits
        | P.Map(int)
        | P.Pipe(enumerate)
        | P.FilterValues()
        | P.Keys()
        | P.Pipe(frozenset)
    )


def intervals_to_bits(intervals: frozenset[int]) -> str:
    bits = ['0'] * 12
    for i in intervals:
        bits[i] = '1'
    return ''.join(bits)


Self = TypeVar('Self', bound='NoteSet')


class NoteSet(Cached, Card):
    """
    an unordered set of notes
    root note is optional
    notes w/o root has no intervals

    hierarchy of classes:
    NoteSet
        NotesWithRoot
            Scale
            Chord
    """
    intervals_to_name: dict[frozenset[int], str] = {}
    name_to_intervals: dict[str, frozenset[int]] = {}

    notes: frozenset[Note]
    root: Note | None
    intervals_ascending: tuple[int, ...] | tuple[()]

    def __init__(
        self,
        notes: frozenset[Note],
        *,
        root: Note | None = None,
    ):
        if not isinstance(notes, frozenset):
            raise TypeError(f'expected frozenset, got {type(notes)}')

        if len(notes) == 0:
            self.notes = frozenset()
        elif typeguards.is_frozenset_of_note(notes):
            self.notes = notes
        else:
            raise TypeError('expected frozenset of Note')

        self.notes_octave_fit = tuple(sorted(self.notes))

        if root is None:
            self.root = root
            self.notes_ascending = self.notes_octave_fit
            self.intervals_ascending = ()
        elif not isinstance(root, Note):
            raise TypeError('root type should be Note | None')
        elif root not in self.notes:
            raise KeyError('root should be one of notes')
        else:
            self.root = root
            root_i = self.notes_octave_fit.index(self.root)
            self.notes_ascending = self.notes_octave_fit[root_i:] + self.notes_octave_fit[:root_i]
            self.intervals_ascending = tuple(note - self.root for note in self.notes_ascending)
        self.intervals = frozenset(self.intervals_ascending)
        self.note_to_interval = dict(zip(self.notes_ascending, self.intervals_ascending))
        self.bits = intervals_to_bits(self.intervals)
        self.name = self.__class__.intervals_to_name.get(self.intervals)

        self.key = self.notes, self.root
        self.note_i = {note: i for i, note in enumerate(self.notes_ascending)}

    @property
    def rootless(self) -> NoteSet:
        return NoteSet(self.notes)

    def with_root(self, root: Note) -> NoteSet:
        if len(self) == 0:
            raise NotImplementedError('cannot add root to empty noteset')
        if not isinstance(root, Note):
            raise TypeError('root type should be Note')
        if root not in self.notes:
            raise KeyError('root should be one of notes')
        return NoteSet(self.notes, root=root)

    def transpose_to(self, note: Note) -> NoteSet:
        if self.root is None:
            raise ValueError('noteset should have root to be transposed')
        return NoteSet.from_intervals(self.intervals, note)

    @classmethod
    def from_name(cls: type[Self], root: str | Note, name: str) -> Self:
        if isinstance(root, str):
            root = Note(root)
        notes = frozenset(root + interval for interval in cls.name_to_intervals[name])
        return cls(notes, root=root)

    @classmethod
    def from_intervals(cls: type[Self], intervals: frozenset[int], root: str | Note | None) -> Self:
        if root is None:
            return cls(frozenset(), root=root)
        if isinstance(root, str):
            root = Note(root)
        return cls(frozenset(root + interval for interval in intervals), root=root)

    @classmethod
    def random(cls: type[Self], n_notes: int | None = None) -> Self:
        if n_notes is None:
            n_notes = random.randint(2, 5)
        notes = frozenset(map(Note, random.sample(config.chromatic_notes, n_notes)))
        return cls(notes)

    @classmethod
    def from_str(cls: type[Self], string: str) -> Self:
        notes, _, root_ = string.partition('/')
        kw = {'root': Note(root_)} if root_ else {}
        return cls(frozenset(Note(note) for note in notes), **kw)

    @overload
    def add_note(self, note: str, steps: int) -> Note | SpecificNote:
        ...

    @overload
    def add_note(self, note: SpecificNote, steps: int) -> SpecificNote:
        ...

    @overload
    def add_note(self, note: Note, steps: int) -> Note:
        ...

    def add_note(self, note: AnyNote, steps: int) -> Note | SpecificNote:
        if len(self) == 0:
            raise NotImplementedError('cannot add notes using empty noteset')
        notes = self.notes_ascending
        if isinstance(note, str):
            note = str_to_note(note)
        if isinstance(note, Note):
            return notes[(notes.index(note) + steps) % len(notes)]
        elif isinstance(note, SpecificNote):
            octaves, i = divmod(self.notes_octave_fit.index(note.abstract) + steps, len(self.notes))
            return SpecificNote(self.notes_octave_fit[i], octave=note.octave + octaves)
        else:
            raise TypeError

    def subtract(self, left: AnyNote, right: AnyNote) -> int:
        if isinstance(left, str):
            left = str_to_note(left)
        if isinstance(right, str):
            right = str_to_note(right)

        if type(left) is type(right) is Note:
            return (self.note_i[left] - self.note_i[right]) % len(self)
        elif type(left) is type(right) is SpecificNote:
            return self.note_i[left.abstract] - self.note_i[right.abstract] + len(self) * (left.octave - right.octave)
        else:
            raise TypeError('left and right should be either Note or SpecificNote (or str representation)')

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            return NotImplemented
        return self.key == other.key

    def __hash__(self) -> int:
        return hash(self.key)

    def __len__(self) -> int:
        return len(self.notes)

    def __getitem__(self, item: int) -> Note:
        return self.notes_ascending[item]

    def __iter__(self) -> Iterator[Note]:
        return iter(self.notes_ascending)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, Note):
            return NotImplemented
        return item in self.notes

    def __le__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            return NotImplemented
        return self.notes <= other.notes

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            return NotImplemented
        return other.notes <= self.notes

    def __repr__(self) -> str:
        x = ''.join(note.name for note in self)
        if self.root is not None:
            return f'{x}/{self.root.name}'
        return x

    def to_piano_image(self) -> str:
        from musiclib.piano import Piano  # hack to fix circular import
        return Piano(note_colors={note: RED for note in self})._repr_svg_()

    def _repr_html_(
        self,
        html_classes: tuple[str, ...] = (),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
    ) -> str:
        return self.repr_card(
            html_classes=html_classes,
            title=title or repr(self),
            subtitle=subtitle,
            header_href=header_href,
            background_color=background_color,
            piano_html=self.to_piano_image(),
        )

    def __getnewargs_ex__(self) -> tuple[tuple[frozenset[Note]], dict[str, Note | None]]:
        return (self.notes,), {'root': self.root}


def subsets(noteset: NoteSet, min_notes: int = 1) -> frozenset[NoteSet]:
    out = set()
    for n_subset in range(min_notes, len(noteset) + 1):
        for notes in itertools.combinations(noteset, n_subset):
            out.add(NoteSet(frozenset(notes)))
    return frozenset(out)



from __future__ import annotations

import functools
from collections.abc import Iterable
from typing import overload

from musiclib import config
from musiclib.util.cache import Cached


@functools.total_ordering
class Note(Cached):
    """
    abstract note, no octave/key
    kinda music theoretic pitch-class
    """

    def __init__(self, name: str):
        """param name: one of CdDeEFfGaAbB"""
        self.name = name
        self.i = config.note_i[name]
        self.is_black = config.is_black[name]

    @classmethod
    def from_i(cls, i: int) -> Note:
        return cls(config.chromatic_notes[i % 12])

    def __repr__(self) -> str:
        return f'Note(name={self.name})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, Note):
            return self.name == other.name
        else:
            return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.i <= config.note_i[other]
        elif isinstance(other, Note):
            return self.i <= other.i
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.name)

    def __add__(self, other: int) -> Note:
        return Note.from_i(self.i + other)

    @overload
    def __sub__(self, other: Note) -> int:
        ...

    @overload
    def __sub__(self, other: int) -> Note:
        ...

    def __sub__(self, other: Note | int) -> int | Note:
        """
        kinda constraint (maybe it will be changed later):
            if you're computing distance between abstract notes - then self considered above other
            G - C == 7 # C0 G0
            C - G == 5 # G0 C1
        """
        if isinstance(other, Note):
            if other.i <= self.i:
                return self.i - other.i
            return 12 + self.i - other.i
        elif isinstance(other, int):
            return self + (-other)

    def __getnewargs__(self) -> tuple[str]:
        return self.name,


@functools.total_ordering
class SpecificNote(Cached):
    def __init__(self, abstract: Note | str, octave: int):
        if isinstance(abstract, str):
            abstract = Note(abstract)
        self.abstract = abstract
        self.is_black = abstract.is_black
        self.octave = octave
        self.i: int = octave * 12 + self.abstract.i  # this is also midi_code
        self._key = self.abstract, self.octave

    @classmethod
    def from_i(cls, i: int) -> SpecificNote:
        div, mod = divmod(i, 12)
        return cls(Note(config.chromatic_notes[mod]), octave=div)

    @classmethod
    def from_str(cls, string: str) -> SpecificNote:
        return cls(Note(string[0]), int(string[1:]))

    def __repr__(self) -> str:
        return f'{self.abstract.name}{self.octave}'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SpecificNote):
            return self._key == other._key
        elif isinstance(other, str):
            return self._key == SpecificNote.from_str(other)._key
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self._key)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SpecificNote):
            return NotImplemented
        return self.i < other.i

    @overload
    def __sub__(self, other: SpecificNote) -> int:
        ...

    @overload
    def __sub__(self, other: int) -> SpecificNote:
        ...

    # @functools.cache
    def __sub__(self, other: SpecificNote | int) -> int | SpecificNote:
        if isinstance(other, SpecificNote):  # distance between notes
            return self.i - other.i
        elif isinstance(other, int):  # subtract semitones
            return self + (-other)
        else:
            raise TypeError(f'SpecificNote.__sub__ supports only SpecificNote | int, got {type(other)}')

    def __add__(self, other: int) -> SpecificNote:
        """C + 7 = G"""
        return SpecificNote.from_i(self.i + other)

    @staticmethod
    def to_abstract(notes: Iterable[SpecificNote]) -> frozenset[Note]:
        return frozenset(note.abstract for note in notes)

    def __getnewargs__(self) -> tuple[Note, int]:
        return self.abstract, self.octave


AnyNote = str | Note | SpecificNote


def str_to_note(note: str) -> Note | SpecificNote:
    if len(note) == 0:
        raise ValueError('invalid note string representation')
    if len(note) == 1:
        return Note(note)
    return SpecificNote.from_str(note)


WHITE_NOTES = frozenset(map(Note, 'CDEFGAB'))
BLACK_NOTES = frozenset(map(Note, 'defab'))



from __future__ import annotations

from collections.abc import Iterator
from collections.abc import Sequence
from typing import overload

from musiclib import config
from musiclib.card import Card
from musiclib.note import SpecificNote
from musiclib.noteset import NoteSet

CHROMATIC_NOTESET = NoteSet.from_str(config.chromatic_notes)


class NoteRange(Sequence[SpecificNote], Card):
    def __init__(
        self,
        start: SpecificNote | str,
        stop: SpecificNote | str,
        noteset: NoteSet = CHROMATIC_NOTESET,
    ):
        if isinstance(start, str):
            start = SpecificNote.from_str(start)
        if isinstance(stop, str):
            stop = SpecificNote.from_str(stop)

        """both ends included"""
        if start > stop:
            raise ValueError('start should be less than stop')

        if not {start.abstract, stop.abstract} <= noteset.notes:
            raise KeyError('start and stop notes should be in the noteset')

        self.start = start
        self.stop = stop
        self.noteset = noteset
        self._key = self.start, self.stop, self.noteset

    def _getitem_int(self, item: int) -> SpecificNote:
        if 0 <= item < len(self):
            q = self.noteset.add_note(self.start, item)
            return q
        elif -len(self) <= item < 0:
            return self.noteset.add_note(self.stop, item + 1)
        else:
            raise IndexError('index out of range')

    @overload
    def __getitem__(self, i: int) -> SpecificNote:
        ...

    @overload
    def __getitem__(self, s: slice) -> NoteRange:
        ...

    def __getitem__(self, item: int | slice) -> SpecificNote | NoteRange:
        if isinstance(item, int):
            return self._getitem_int(item)
        elif isinstance(item, slice):
            # if item.start is None:
            start = 0 if item.start is None else item.start
            stop = len(self) - 1 if item.stop is None else item.stop
            if not 0 <= start <= stop <= len(self):
                raise IndexError('NoteRange slice is out of range, negative indexing is not supported')
            return NoteRange(self._getitem_int(start), self._getitem_int(stop), self.noteset)
        else:
            raise TypeError(f'NoteRange indices must be integers or slices, got {type(item)}')

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, SpecificNote):
            return False
        return item.abstract in self.noteset and self.start <= item <= self.stop

    def __iter__(self) -> Iterator[SpecificNote]:
        return (self[i] for i in range(len(self)))

    def __repr__(self) -> str:
        return f'NoteRange({self.start}, {self.stop}, noteset={self.noteset})'

    def __len__(self) -> int:
        return self.noteset.subtract(self.stop, self.start) + 1

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoteRange):
            return NotImplemented
        return self._key == other._key

    def __hash__(self) -> int:
        return hash(self._key)

    def to_piano_image(self) -> str:
        from musiclib.piano import Piano  # hack to fix circular import
        return Piano(
            note_colors=None if self.noteset is CHROMATIC_NOTESET else dict.fromkeys(self.noteset, config.RED),
            squares={self.start: {'text': str(self.start), 'text_size': '8'}, self.stop: {'text': str(self.stop), 'text_size': '8'}},
            noterange=NoteRange(self.start, self.stop),
        )._repr_svg_()

    def _repr_html_(
        self,
        html_classes: tuple[str, ...] = ('card',),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
    ) -> str:
        return self.repr_card(
            html_classes=html_classes,
            title=title or f'NoteRange({self.start}, {self.stop})',
            subtitle=subtitle,
            header_href=header_href,
            background_color=background_color,
            piano_html=self.to_piano_image(),
        )



from __future__ import annotations

import itertools
import pickle
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Sequence
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from typing import Any
from typing import TypeVar

import tqdm

Op = TypeVar('Op')

OpsIterable = Iterable[Op]
OpsFixedPerStep = Sequence[Iterable[Op]]
OpsCallableFromCurr = Callable[[Op], Iterable[Op]]


OPTIONS_EXCEPTION = TypeError('options, options_i, options_callable are mutually exclusive. Only 1 must be not None')


class SequenceBuilder:
    def __init__(  # noqa: C901
        self,
        n: int,
        options: OpsIterable[Op] | None = None,
        options_i: OpsFixedPerStep[Op] | None = None,
        options_callable: OpsCallableFromCurr[Op] | None = None,
        curr_prev_constraint: dict[int, Callable[[Op, Op], bool]] | None = None,
        candidate_constraint: Callable[[tuple[Op, ...]], bool] | None = None,
        i_constraints: dict[int, Callable[[Op], bool]] | None = None,
        unique_key: Callable[[Op], Any] | None = None,
        loop: bool = False,
        prefix: tuple[Op, ...] = (),
        parallel: bool = False,
    ):
        if parallel:
            try:
                pickle.dumps(options)
                pickle.dumps(options_i)
                pickle.dumps(options_callable)
                pickle.dumps(curr_prev_constraint)
                pickle.dumps(candidate_constraint)
                pickle.dumps(i_constraints)
                pickle.dumps(unique_key)
            except BaseException:
                raise pickle.PickleError('parallel=True requires all arguments should be picklable')

        self.n = n
        self.options: tuple[Op, ...] | None
        self.options_i: OpsFixedPerStep[Op] | None
        self.options_callable: OpsCallableFromCurr[Op] | None

        if options is not None:
            if not (options_i is None and options_callable is None):
                raise OPTIONS_EXCEPTION
            options_seq = tuple(options)
            if len(options_seq) != len(frozenset(options_seq)):
                raise ValueError('options should be unique')
            self.options = options_seq
            self.generate_options = self._generate_options_iterable
        elif options_i is not None:
            if not (options is None and options_callable is None):
                raise OPTIONS_EXCEPTION
            if len(options_i) != n:
                raise ValueError('options_i should have options for each step up to n')
            self.options = options
            self.generate_options = self._generate_options_fixed_per_step
        elif options_callable is not None:
            if not (options is None and options_i is None):
                raise OPTIONS_EXCEPTION
            self.options = options
            self.generate_options = self._generate_options_callable_from_curr
        self.options_i = options_i
        self.options_callable = options_callable

        if curr_prev_constraint is not None and not all(k < 0 for k in curr_prev_constraint.keys()):
            raise KeyError('curr_prev_constraint keys must be negative')
        self.curr_prev_constraint = curr_prev_constraint
        self.candidate_constraint = candidate_constraint
        self.i_constraints = i_constraints
        self.unique_key = unique_key
        self.loop = loop
        self.prefix = prefix
        self.parallel = parallel

    def _generate_options_iterable(self, seq: tuple[Op, ...]) -> Iterable[Op]:
        if self.options is not None:
            return self.options
        raise TypeError

    def _generate_options_fixed_per_step(self, seq: tuple[Op, ...]) -> Iterable[Op]:
        if self.options_i is not None:
            return self.options_i[len(seq)]
        raise TypeError

    def _generate_options_callable_from_curr(self, seq: tuple[Op, ...]) -> Iterable[Op]:
        if self.options_callable is not None:
            return self.options_callable(seq[-1])
        raise TypeError

    def __iter__(self) -> Generator[tuple[Op, ...], None, None]:
        return self._iter(self.prefix)

    def _iter(self, prefix: tuple[Op, ...] = ()) -> Generator[tuple[Op, ...], None, None]:
        seq = prefix or ()
        ops = self.generate_options(seq)

        if self.i_constraints is not None and (i_constraint := self.i_constraints.get(len(seq))):
            ops = filter(i_constraint, ops)

        if self.unique_key:
            prefix_keys = frozenset(self.unique_key(op) for op in prefix)
            ops = [op for op in ops if self.unique_key(op) not in prefix_keys]

        if self.curr_prev_constraint:

            if abs(max(self.curr_prev_constraint.keys())) >= self.n:
                raise IndexError('max index to look back in curr_prev_constraint should be less than n')

            for k, f in self.curr_prev_constraint.items():
                if abs(k) > len(seq):
                    continue
                ops = [op for op in ops if f(seq[k], op)]
        ops = tuple(ops)
        map_func = partial(self._generate_candidates, seq=seq)

        if len(prefix) == len(self.prefix):
            if self.parallel:
                with ProcessPoolExecutor() as executor:
                    for it in tqdm.tqdm(executor.map(map_func, ops), total=len(ops)):
                        yield from it
                return
            else:
                it = tqdm.tqdm(map(map_func, ops), total=len(ops))
        else:
            it = map(map_func, ops)
        it = itertools.chain.from_iterable(it)
        yield from it

    def _generate_candidates(self, op: Op, seq: tuple[Op, ...]) -> Iterable[tuple[Op, ...]]:
        def inner() -> Iterable[tuple[Op, ...]]:
            candidate = seq + (op,)
            if self.candidate_constraint is not None and not self.candidate_constraint(candidate):
                return
            if len(candidate) < self.n:
                yield from self._iter(prefix=candidate)
                return
            if self.curr_prev_constraint and self.loop and not all(
                f(candidate[(i + k) % self.n], candidate[i])
                for k, f in self.curr_prev_constraint.items()
                    for i in range(abs(k))
            ):
                return
            yield candidate
        out = tuple(inner())
        return out



from collections.abc import Hashable
from typing import Any


class Cached:
    _cache: dict[tuple[type, tuple[Hashable, ...], frozenset[tuple[str, Hashable]]], Any] = {}

    def __new__(cls, *args: Hashable, **kwargs: Hashable) -> Any:
        key = cls, args, frozenset(kwargs.items())
        instance = cls._cache.get(key)
        if instance is not None:
            return instance
        instance = super().__new__(cls)
        cls._cache[key] = instance
        return instance



import asyncio
import functools
from collections.abc import Iterable
from pathlib import Path

import mido

from musiclib.chord import SpecificChord
from musiclib.note import SpecificNote
from musiclib.rhythm import Rhythm

Playable = SpecificNote | SpecificChord


class Player:
    def __init__(self, midi_device: str | None = None) -> None:
        if midi_device is None:
            self.send_message = self._print_message
        else:
            self.port = mido.open_output(midi_device)
            self.send_message = self._send_message

    def _print_message(self, *args: str | int, note: int, **kwargs: str | int) -> None:
        print('MIDI_DEVICE not found |', *args, f'{note=},', ', '.join(f'{k}={v!r}' for k, v in kwargs.items()))

    def _send_message(self, *args: str | int, note: int, **kwargs: str | int) -> None:
        note += 24  # to match ableton octaves
        self.port.send(mido.Message(*args, note=note, **kwargs))

    @functools.singledispatchmethod
    async def play(self, obj: Playable, seconds: float = 1) -> None:
        ...

    @play.register
    async def _(self, obj: SpecificNote, seconds: float = 1) -> None:
        self.send_message('note_on', note=obj.i, channel=0)
        await asyncio.sleep(seconds)
        self.send_message('note_off', note=obj.i, channel=0)

    @play.register
    async def _(self, obj: SpecificChord, seconds: float = 1, bass_octave: int | None = None) -> None:
        tasks = [self.play(note, seconds) for note in obj.notes]
        if bass_octave:
            if obj.root is None:
                raise ValueError('cannot play bass when root is None')
            tasks.append(self.play(SpecificNote(obj.root, bass_octave), seconds))
        await asyncio.gather(*tasks)


def chord_to_midi(
    chord: SpecificChord,
    path: str | Path | None = None,
    n_bars: int = 1,
) -> mido.MidiFile | None:
    mid = mido.MidiFile(type=0, ticks_per_beat=96)
    track = mido.MidiTrack()
    track.append(mido.MetaMessage(type='track_name', name='test_name'))
    track.append(mido.MetaMessage(type='time_signature', numerator=4, denominator=4, clocks_per_click=36))
    track.append(mido.MetaMessage(type='time_signature', numerator=4, denominator=4, clocks_per_click=36))
    if chord.root is not None:
        track.append(mido.MetaMessage(type='marker', text=chord.root.name))

    stop_time = int(n_bars * mid.ticks_per_beat * 4)

    for note in chord.notes_ascending:
        track.append(mido.Message('note_on', note=note.i, velocity=100, time=0))
    for i, note in enumerate(chord.notes_ascending):
        track.append(mido.Message('note_off', note=note.i, velocity=100, time=stop_time if i == 0 else 0))

    mid.tracks.append(track)
    mid.meta = {'chord': chord}
    if path is None:
        return mid
    mid.save(path)
    return None


TO_MIDI_MUTUAL_EXCLUSIVE_ERROR = TypeError('note_, chord are mutually exclusive. Only 1 must be not None')


def rhythm_to_midi(  # noqa: C901
    rhythm: Rhythm,
    path: str | Path | None = None,
    note_: SpecificNote | None = None,
    chord: SpecificChord | None = None,
    progression: Iterable[SpecificChord] | None = None,
) -> mido.MidiFile | None:

    if note_ is not None and chord is not None:
        raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR

    if chord is None:
        if note_ is None:
            raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR
        note__ = note_

    if note_ is None:
        if chord is None:
            raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR

    mid = mido.MidiFile(type=0, ticks_per_beat=96)

    ticks_per_note = mid.ticks_per_beat * rhythm.beats_per_bar // rhythm.bar_notes
    track = mido.MidiTrack()
    track.append(mido.MetaMessage(type='track_name', name='test_name'))
    track.append(mido.MetaMessage(type='time_signature', numerator=4, denominator=4, clocks_per_click=36))
    t = 0

    def append_bar(chord: SpecificChord | None) -> None:
        nonlocal t
        for is_play in rhythm.notes:
            if is_play:
                notes = [note__.i] if chord is None else [note.i for note in chord.notes]
                for i, note in enumerate(notes):
                    track.append(mido.Message('note_on', note=note, velocity=100, time=t if i == 0 else 0))
                for i, note in enumerate(notes):
                    track.append(mido.Message('note_off', note=note, velocity=100, time=ticks_per_note if i == 0 else 0))
                t = 0
            else:
                t += ticks_per_note

    if progression is None:
        append_bar(chord)
    else:
        for chord in progression:
            append_bar(chord)

    mid.tracks.append(track)
    if path is None:
        return mid
    mid.save(path)
    return None



import math

from musiclib.note import SpecificNote


class Pitch:
    def __init__(
        self,
        hz_tuning: float = 440,
        origin_note: SpecificNote = SpecificNote('A', 4),
        transpose: float = 0,
    ):
        """
        origin_note: in midi format, A4 midi ~ A4 ableton ~ 440Hz
        """
        self.hz_tuning = hz_tuning
        self.origin_note = origin_note
        self.transpose = transpose

    def i_to_hz(self, i: float) -> float:
        return self.hz_tuning * 2 ** ((i + self.transpose) / 12)

    def hz_to_i(self, hz: float) -> float:
        return 12 * math.log2(hz / self.hz_tuning) - self.transpose

    def note_i_to_hz(self, note_i: float) -> float:
        return self.i_to_hz(note_i - self.origin_note.i)

    def hz_to_note_i(self, hz: float) -> float:
        return self.origin_note.i + self.hz_to_i(hz)

    def note_to_hz(self, note: SpecificNote) -> float:
        return self.note_i_to_hz(note.i)

    def hz_to_note(self, hz: float) -> SpecificNote:
        return SpecificNote.from_i(round(self.hz_to_note_i(hz)))

    @staticmethod
    def hz_to_px(hz: float, hz_min: float, hz_max: float, px_max: float) -> float:
        """Convert Hz to pixel position (using logarithmic scale)"""
        return math.log2(hz / hz_min) / math.log2(hz_max / hz_min) * px_max

    @staticmethod
    def px_to_hz(px: float, hz_min: float, hz_max: float, px_max: float) -> float:
        """Convert pixel position to hz (assuming pixel using logarithmic scale)"""
        c = px / px_max
        return hz_min ** (1 - c) * hz_max ** c  # type: ignore



import abc


class Card(abc.ABC):
    @staticmethod
    def repr_card(
        html_classes: tuple[str, ...] = (),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
        piano_html: str | None = None,
    ) -> str:
        out = ''
        if title is not None:
            out += f"<div style='font-size: 1em; font-weight: bold' class='card_title'>{title}</div>\n"
        if subtitle is not None:
            out += f"<div style='margin-top: -0.2em;font-size: 0.8em' class='card_subtitle'>{subtitle}</div>\n"
        out = f'''
        <div
            class='card_header'
            style='
                height: 32px;
                font-family: sans-serif;
            '
        >
        {out}
        </div>
        '''

        if header_href is not None:
            out = f'''
            <a href='{header_href}'>
            {out}
            </a>
            '''
        if piano_html is not None:
            out = f'''
            {out}
            {piano_html}
            '''

        background_color = f'background-color: {background_color};' if background_color is not None else ''

        classes = ' '.join(('card', *html_classes))
        out = f'''
        <div
            class='{classes}'
            style='
                margin: 5px;
                width: fit-content;
                padding: 0px 2px 0px 2px;
                border: 1px solid rgba(0,0,0,0.5);
                height: 120px;
                box-shadow: 2px 2px;
                border-radius: 3px;
                {background_color}
            '
        >
        {out}
        </div>
        '''
        return out

    @abc.abstractmethod
    def _repr_html_(
        self,
        html_classes: tuple[str, ...] = (),
        title: str | None = None,
        subtitle: str | None = None,
        header_href: str | None = None,
        background_color: str | None = None,
    ) -> str:
        ...



from __future__ import annotations

import functools
import itertools
import random
import statistics

import pipe21 as P

from musiclib import config
from musiclib.util.cache import Cached
from musiclib.util.sequence_builder import SequenceBuilder


class Rhythm(Cached):
    def __init__(
        self,
        notes: tuple[int, ...],
        beats_per_minute: int = config.beats_per_minute,
        beats_per_bar: int = 4,
        bar_notes: int = 16,  # kinda grid size
    ):
        self.notes = notes
        self.beats_per_minute = beats_per_minute
        self.beats_per_second = beats_per_minute / 60
        self.beats_per_bar = beats_per_bar
        self.bar_seconds = beats_per_bar / self.beats_per_second
        self.bar_notes = bar_notes
        self.note_seconds = self.bar_seconds / bar_notes
        self.bits = ''.join(map(str, self.notes))

    @classmethod
    def random_rhythm(cls, n_notes: int | None = None, bar_notes: int = 16) -> Rhythm:
        if n_notes is None:
            n_notes = random.randint(1, bar_notes)
        if not (0 < n_notes <= bar_notes):
            raise ValueError(f'number of notes should be more than 1 and less than bar_notes={bar_notes}')
        notes = [1] * n_notes + [0] * (bar_notes - n_notes)
        random.shuffle(notes)
        return cls(tuple(notes), bar_notes=bar_notes)

    def __repr__(self) -> str:
        return f"Rhythm('{self.bits}')"

    @functools.cached_property
    def has_contiguous_ones(self) -> bool:
        return (
            self.notes[0] == 1 and self.notes[-1] == 1
            or any(len(list(g)) > 1 for k, g in itertools.groupby(self.notes, key=bool) if k)
        )

    @staticmethod
    def no_contiguous_ones(prev: int, curr: int) -> bool:
        return not (prev == curr == 1)

    @functools.cached_property
    def score(self) -> float:
        """spacings variance"""
        first_1 = self.notes.index(1)
        x = self.notes[first_1:] + self.notes[:first_1]
        spacings = [len(list(g)) for k, g in itertools.groupby(x, key=bool) if not k]
        if len(spacings) == 1:  # TODO: try normalize into 0..1
            return float('inf')
        return statistics.variance(spacings)

    @staticmethod
    def all_rhythms(n_notes: int | None = None, bar_notes: int = 16, sort_by_score: bool = False) -> tuple[Rhythm, ...]:
        rhythms__ = SequenceBuilder(
            n=bar_notes,
            options=(0, 1),
            curr_prev_constraint={-1: Rhythm.no_contiguous_ones},
        )

        if n_notes is not None:
            rhythms_ = rhythms__ | P.Filter(lambda r: sum(r) == n_notes)

        rhythms = (Rhythm(r, bar_notes=bar_notes) for r in rhythms_)

        if sort_by_score:
            rhythms = (
                rhythms
                | P.KeyBy(lambda rhythm: rhythm.score)
                | P.Pipe(lambda x: sorted(x, key=lambda score_rhythm: (score_rhythm[0], score_rhythm[1].notes)))
            )
        out: tuple[Rhythm] = rhythms | P.Pipe(tuple)
        return out



from __future__ import annotations

import functools
import itertools
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence
from typing import overload

from musiclib.chord import SpecificChord
from musiclib.note import SpecificNote
from musiclib.util.cache import Cached

CheckCallable = Callable[[SpecificChord, SpecificChord], bool]


class Progression(Cached, Sequence[SpecificChord]):
    def __init__(self, chords: tuple[SpecificChord, ...], /):
        if not all(isinstance(x, SpecificChord) for x in chords):
            raise TypeError('only SpecificChord items allowed')
        self.chords = chords

    @overload
    def __getitem__(self, i: int) -> SpecificChord:
        ...

    @overload
    def __getitem__(self, s: slice) -> Progression:
        ...

    def __getitem__(self, item: int | slice) -> SpecificChord | Sequence[SpecificChord]:
        if isinstance(item, slice):
            return Progression(self.chords[item])
        return self.chords[item]

    def __iter__(self) -> Iterator[SpecificChord]:
        return iter(self.chords)

    def __len__(self) -> int:
        return len(self.chords)

    def __repr__(self) -> str:
        return f'Progression{self.chords}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Progression):
            return NotImplemented
        return self.chords == other.chords

    def __hash__(self) -> int:
        return hash(self.chords)

    def all(self, checks__: Iterable[CheckCallable]) -> bool:
        return all(check(a, b) for a, b in itertools.pairwise(self) for check in checks__)

    def all_not(self, checks__: Iterable[CheckCallable]) -> bool:
        return all(not check(a, b) for a, b in itertools.pairwise(self) for check in checks__)

    @property
    def distance(self) -> int:
        n = len(self)
        return sum(abs(self[i] - self[(i + 1) % n]) for i in range(n))

    def transpose_unique_key(self, origin_name: bool = True) -> tuple[frozenset[int], ...] | tuple[int, tuple[frozenset[int], ...]]:
        origin = self[0][0]
        key = tuple(frozenset(note - origin for note in chord.notes) for chord in self)
        if origin_name:
            return origin.abstract.i, key
        return key

    def __add__(self, other: int) -> Progression:
        if not isinstance(other, int):
            raise TypeError('only adding integers is allowed (transposition)')
        return Progression(tuple(chord + other for chord in self))

    @functools.cached_property
    def transposed_to_C0(self) -> Progression:
        return self + (SpecificNote('C', 0) - self[0][0])

    def __getnewargs__(self) -> tuple[tuple[SpecificChord, ...]]:
        return self.chords,



from musiclib.chord import Chord
from musiclib.scale import Scale


def str_to_chord_progression(s: Scale, progression: str) -> tuple[Chord, ...]:
    return tuple(s.triads[s.notes_ascending.index(c)] for c in progression)
